from collections import defaultdict
import json
import os

def pad(seconds):
	if seconds < 10:
		return "0" + str(seconds)
	return str(seconds)

def clocktime(loop):
	seconds = loop/16
	return str(seconds/60) + ":" + pad(seconds%60)

cycles = {}
cycles['CommandCenter'] = 100
cycles['SCV'] = 17
cycles['OrbitalCommand'] = 35
cycles['OrbitalCommandFlying'] = 0
cycles['CommandCenterFlying'] = 0
cycles['PlanetaryFortress'] = 50
cycles['Marine'] = 25
cycles['Marauder'] = 30
cycles['Reaper'] = 45

cycles['Hellion'] = 30
cycles['BattleHellion'] = 30
cycles['WidowMine'] = 40
cycles['SiegeTank'] = 45

cycles['Medivac'] = 42
cycles['Raven'] = 60
cycles['VikingFighter'] = 42

cycles['Zealot'] = 38
cycles['Stalker'] = 42
cycles['Sentry'] = 37

cycles['Probe'] = 17
cycles['Immortal'] = 55
cycles['Colossus'] = 75


cycles['VoidRay'] = 60
cycles['Phoenix'] = 30
cycles['Oracle'] = 50

cycles['Drone'] = 17
cycles['Queen'] = 50
cycles['Zergling'] = 24
cycles['Hydralisk'] = 33
cycles['Roach'] = 27
cycles['Overlord'] = 25


cycles['Barracks'] = 65

for name in cycles:
	cycles[name] = cycles[name]*16-1

canCreate = {}
canCreate['CommandCenter'] = ['SCV']
canCreate['PlanetaryFortress'] = ['SCV']
canCreate['OrbitalCommand'] = ['SCV']

canCreate['OrbitalCommandFlying'] = []
canCreate['CommandCenterFlying'] = []

canTransformInto = {}
canTransformInto['CommandCenter'] = ['PlanetaryFortress','OrbitalCommand','CommandCenterFlying']

class Unit:
	def __init__(self,name,tag):
		self.name=name
		self.tag = tag

class Building:
	def __init__(self,name,done,tag,loop,owner):
		self.name = name
		self.done=done
		self.idle=0
		self.tag = tag
		self.idle_periods = []
		self.owner = owner
		self.last_used = loop

	def could_have_done(self,tag,element,loop,alteration,players):
		if alteration and element not in canTransformInto[self.name]:
			return False
		if not alteration and element not in canCreate[self.name]:
			return False

		if loop-cycles[element] >= self.last_used:
			idle_amount = loop-(self.last_used+cycles[element])

			if idle_amount > 0:
				idle_period = (loop-cycles[element]-idle_amount,loop-cycles[element])
#				print clocktime(loop),players[self.owner].name,"idle",self.name,self.tag,":",idle_period,"which makes for",idle_amount/16.0,"seconds, from",clocktime(idle_period[0]),"to",clocktime(idle_period[1])
				self.idle_periods.append(idle_period)

				self.idle += loop-cycles[element]-self.last_used

			self.last_used = loop
			return True

		return False

	def alter(self,tag,newAlteration,loop,players):
		if self.could_have_done(tag,newAlteration,loop,True,players):
			self.name=newAlteration
			return True
#		print "Cannot alter from",self.name,"to",newAlteration
		return False

	def getProductionLines(self):
		if self.name.endswith('Flying'):
			return 0
		return 1

	def setCompleted(self,loop):
		self.last_used=loop
		self.done=True

	def setFlying(self,loop):
		self.flying = True

	def setNotFlying(self,loop):
		self.flying = False

	def createUnit(self,tag,unit,loop,players):
		if not self.done:
			return False

		if self.flying:
			return False

		return self.could_have_done(self,tag,unit,loop,players)

class Player:
	def __init__(self,name,race):
		self.buildings = {}
		self.units = {}
		self.name = name.encode('utf-8')
		self.race = race
		self.dead_buildings = []

	def addBuilding(self,building):
		self.buildings[building.tag] = building

	def addUnit(self,unitName,tag,loop,players):
		# Assign to Command Centers last because of the possibility
		# of it being upgraded instead of used to make workers
		for building in self.buildings:
			if self.buildings[building].name != 'CommandCenter':
				if self.buildings[building].could_have_done(tag,unitName,loop,False,players):
					self.units[tag] = Unit(unitName,tag)
					return True

		for building in self.buildings:
			if self.buildings[building].could_have_done(tag,unitName,loop,False,players):
				self.units[tag] = Unit(unitName,tag)
				return True

		if loop > 0:
#			print clocktime(loop),"Could not find anything that could have made that",unitName
			return False

		self.units[tag] = Unit(unitName,tag)
		return True

	def alterBuilding(self,tag,newAlteration,loop,players):
		if self.buildings[tag].alter(tag,newAlteration,loop,players):
#			print clocktime(loop),self.name,"now has an altered building, new look is",newAlteration
			return True
		return False

	def ownsBuildingTag(self,tag):
		return tag in self.buildings

	def ownsUnitTag(self,tag):
		return tag in self.units

	def loseUnit(self,tag):
		del self.units[tag]

	def loseBuilding(self,tag):
		self.dead_buildings.append(self.buildings[tag])
		del self.buildings[tag]

def add_idleness(idle,building):
	for period in building.idle_periods:
		for i in range(period[0],period[1]):
			idle[i] += 1

class ProductionInfo():
	def __init__(self):
		pass

	def getCycles(self,unit):
		try:
			return cycles[unit]
		except KeyError:
			print "*** I do not know how many cycles a",unit,"takes ***"

class ProductionLog:
	def __init__(self):
		self.active = defaultdict(lambda: defaultdict(int))
		self.available = defaultdict(lambda: defaultdict(int))
		self.inProduction = defaultdict(lambda: defaultdict(int))

	def addUnit(self,unit,completedAt):
		if completedAt == 0:
			return

		for frame in range(completedAt-cycles[unit], completedAt):
			self.inProduction[self.findBuilding(unit)][unit] += 1
			self.active[self.findBuilding(unit)][frame] += 1

	def addUpgrade(self,upgrade,completedAt):
		if completedAt == 0:
			return

		for frame in range(completedAt-cycles[upgrade], completedAt):
			self.inProduction[self.findUpgradeBuilding(upgrade)][upgrade] += 1
			self.active[self.findUpgradeBuilding(upgrade)][frame] += 1

	def getBaseName(self,building):
		if building == 'OrbitalCommand' or building == 'PlanetaryFortress':
			return 'CommandCenter'
		return building

	def getProductionLines(self,building):
		if building.endswith('Reactor'):
			return 2
		if building.endswith('Flying'):
			return 0
		return 1
			
	def addBuilding(self,building,completedAt):
		print "Adding",building
		for frame in range(completedAt,completedAt+16*60*60):
			self.available[self.getBaseName(building)][frame] += self.getProductionLines(building)

# This is optional and low-priority.
	def findUpgradeBuilding(self,upgrade):
		if upgrade in ['WarpGateResearch']:
			return 'CyberneticsCore'

		elif upgrade in ['ShieldWall','Stimpack','PunisherGrenades',]:
			return 'BarracksTechLab'

		elif upgrade in []:
			return 'FactoryTechLab'

		#  'DurableMaterials' == Increases the duration of Auto-Turrets and Point Defense Drones from 180 to 240 seconds.
		elif upgrade in ['MedivacCaduceusReactor','DurableMaterials','RavenCorvidReactor', 'BansheeCloak']: 
			return 'StarportTechLab'

		elif upgrade in ['TerranInfantryWeaponsLevel1','TerranInfantryArmorsLevel1','HiSecAutoTracking', 'NeosteelFrame','TerranBuildingArmor',]:
			return 'EngineeringBay'

		elif upgrade in ['TerranVehicleAndShipWeaponsLevel1', 'TerranVehicleAndShipArmorsLevel1',]:
			return 'Armory'


	def findBuilding(self,unit):
		if unit in ['SCV']:
			return 'CommandCenter'

		elif unit in ['Marine', 'Reaper', 'Marauder']:
			return 'Barracks'
		
		elif unit in ['Medivac', 'Viking', 'VikingFighter', 'Raven', 'Battlecruiser']:
			return 'Starport'

		elif unit in ['Hellion', 'WidowMine', 'BattleHellion', 'SiegeTank']:
			return 'Factory'

		elif unit in ['Probe']:
			return 'Nexus'

		elif unit in ['Oracle', 'Phoenix', 'VoidRay']:
			return 'Stargate'

		elif unit in ['Zealot', 'Stalker', 'Sentry', 'HighTemplar', 'DarkTemplar']:
			return 'Gateway'

		elif unit in ['Colossus', 'Immortal', 'Observer', 'WarpPrism']:
			return 'RoboticsFacility'


		# Zerg units are special
		elif unit in ['Roach', 'Hydralisk', 'Mutalisk', 'Drone', 'Overlord', 'Viper', 'Corruptor', 'Zergling']:
			return 'Larva'

		elif unit in ['Queen', 'Larva']:
			return 'Hatchery'

		return 'Unknown building [' + unit + ']'

	def idleReport(self):
		idle = defaultdict(lambda: defaultdict(int))
		for building in self.available:
			for frame in self.available[building]:
				idle[building][frame] = self.available[building][frame] - self.active[building][frame]

		return idle

	def __str__(self):
		return "Active: " + str(self.active) + "\nAvailable: " + str(self.available)

	def __unicode__(self):
		return self.__str__()

class IdleProductionAnalyzer:
	def __init__(self,replayfile):
		from replayparsing import Replay
		self.replay = Replay(replayfile)
		self.analyze()
		self.worker_death_stats = self.create_worker_death_report()
		(self.lost, self.cancelled) = self.create_lost_cancelled_report()

	def analyze(self):
		replay = self.replay

		self.productionLog = {}
		created = {}

		players = {}
		for player in replay.players:
			self.productionLog[player] = ProductionLog()
			created[player] = defaultdict(int)
			players[player] = Player(replay.players[player]['name'], replay.players[player]['race'])

		for event in replay.trackerEvents:
			try:
				player = event.get('m_controlPlayerId',None)

				loopNumber = event['_gameloop']
				
				if event.get('_event','') == 'NNet.Replay.Tracker.SUnitTypeChangeEvent':
					tag = event['m_unitTagIndex']
					
					for P in players:
						if players[P].ownsBuildingTag(tag):
							if not players[P].alterBuilding(tag,event['m_unitTypeName'],loopNumber,players):
								print "Unsuccessfully altered building into",event['m_unitTypeName'],event
							
				elif event.get('_event','') == 'NNet.Replay.Tracker.SUnitDiedEvent':
					for P in players:
						if players[P].ownsBuildingTag(tag):
							players[P].loseBuilding(tag)
						elif players[P].ownsUnitTag(tag):
							players[P].loseUnit(tag)
								
				elif event.get('_event','') == 'NNet.Replay.Tracker.SUnitBornEvent':
					tag = event['m_unitTagIndex']
					unit = event['m_unitTypeName']
					owner = event['m_controlPlayerId']

					if owner == 0:
						print "OWNER ZERO:",event
						continue

					self.productionLog[owner].addUnit(unit,loopNumber)
								
					if unit == 'CommandCenter':
						players[event['m_controlPlayerId']].addBuilding(Building(unit,True,event['m_unitTagIndex'],loopNumber,event['m_controlPlayerId']))
									
					elif unit == 'SCV':
						players[event['m_controlPlayerId']].addUnit(unit,tag,loopNumber,players)

					if event.get('_gameloop', -1) > 0 or unit in ['SCV', 'Drone', 'Probe', 'Larva']:
						created[event['m_controlPlayerId']][event['m_unitTypeName']] += 1
										
				elif event.get('_event','') == 'NNet.Replay.Tracker.SUnitInitEvent':
					players[player].addBuilding(Building(event['m_unitTypeName'],False,event['m_unitTagIndex'],loopNumber,event['m_controlPlayerId']))

				elif event.get('_event', '') == 'NNet.Replay.Tracker.SUnitDoneEvent':
					tag = event['m_unitTagIndex']
					
					for P in players:	
						if tag in players[P].buildings:
							self.productionLog[owner].addBuilding(players[P].buildings[tag].name, loopNumber)
							players[P].buildings[tag].setCompleted(loopNumber)

				elif event.get('_event', '') == 'NNet.Replay.Tracker.SUpgradeEvent':
					for P in players:
						if players[P].ownsBuildingTag(tag):
							pass # increase counter here
												
			except KeyError as keyError:
				import traceback
				traceback.print_exc()
				print "KeyError",event,keyError

		self.report = self.create_report(replay,players.values(),replay.frames)

					
	def create_report(self,replay,players,game_length_in_frames):
		report = ""

		report += "var mapname = '" + replay.game_type + ' on ' + replay.map + "';\n "
		
		report += 'var series = [\n'
		
		for player in players:
#			if player.race != 'Terran':
#				continue


			playerName = player.name
			report += '{\nname: \''


			report += "".join([x if ord(x) < 128 else '?' for x in playerName])
			report += ' (' + player.race + ')\',\n'
			
			INCREMENT = 60 # seconds
			
			increments = defaultdict(int)
			
			idle = defaultdict(int)
			for building in player.buildings:
				add_idleness(idle,player.buildings[building])

			for building in player.dead_buildings:
				add_idleness(idle,building)
					
			for frame in idle:
				increments[frame/16/INCREMENT]+=1

			for increment in increments:
				increments[increment]/=16
							
			result = []
			total = 0
			for increment in xrange(game_length_in_frames/16/INCREMENT):
				result.append(total)
				total += increments[increment]

			if player.race == 'Terran':
				report += '\t\t\t\tdata: ' + str(result) + ',\n'
			report += '\t\t\t},'
								
		report += '];'
		return report

	def create_worker_death_report(self):
		killed = {}

		for player in self.replay.players:
			killed[player] = {}

		for player in self.replay.players:
			for entry in self.replay.players[player]['unitsLost']:
				if entry['killer'] not in killed:
					killed[entry['killer']] = {}

				k = killed[entry['killer']].get(entry['name'],0) + 1
				killed[entry['killer']][entry['name']] = k
		return killed

	def create_lost_cancelled_report(self):
		lost = {}
		cancelled = {}

		for player in self.replay.players:
			lost[player] = {}
			cancelled[player] = {}

		for player in self.replay.players:
			for entry in self.replay.players[player]['unitsLost']:
				debug = open('debugEvents.txt', 'w')

				debug.write(str(entry) + "\n")

				if entry['killer'] not in self.replay.players:
					k = cancelled[player].get(entry['name'],0) + 1
					cancelled[player][entry['name']] = k

				k = lost[player].get(entry['name'],0) + 1
				lost[player][entry['name']] = k

		return (lost,cancelled)

# Testing
#	ReplayAnalyzer('test.SC2Replay')
